(function (window) {
    'use strict';

    const IDBKeyRange = window.IDBKeyRange || window.webkitIDBKeyRange;
    const transactionModes = {
        readonly: 'readonly',
        readwrite: 'readwrite'
    };
    const hasOwn = Object.prototype.hasOwnProperty;
    const defaultMapper = x => x;

    let indexedDB = (function () {
        if (!indexedDB) {
            indexedDB = window.indexedDB || window.webkitIndexedDB || window.mozIndexedDB || window.oIndexedDB || window.msIndexedDB || window.shimIndexedDB;

            if (!indexedDB) {
                throw new Error('IndexedDB required');
            }
        }
        return indexedDB;
    })();

    const dbCache = {};
    const isArray = Array.isArray;

    function mongoDBToKeyRangeArgs (opts) {
        var keys = Object.keys(opts).sort();
        if (keys.length === 1) {
            var key = keys[0];
            var val = opts[key];
            var name, inclusive;
            switch (key) {
            case 'eq': name = 'only'; break;
            case 'gt':
                name = 'lowerBound';
                inclusive = true;
                break;
            case 'lt':
                name = 'upperBound';
                inclusive = true;
                break;
            case 'gte': name = 'lowerBound'; break;
            case 'lte': name = 'upperBound'; break;
            default: throw new TypeError('`' + key + '` is not valid key');
            }
            return [name, [val, inclusive]];
        }
        var x = opts[keys[0]];
        var y = opts[keys[1]];
        var pattern = keys.join('-');

        switch (pattern) {
        case 'gt-lt': case 'gt-lte': case 'gte-lt': case 'gte-lte':
            return ['bound', [x, y, keys[0] === 'gt', keys[1] === 'lt']];
        default: throw new TypeError(
          '`' + pattern + '` are conflicted keys'
        );
        }
    }
    function mongoifyKey (key) {
        if (key && typeof key === 'object' && !(key instanceof IDBKeyRange)) {
            let [type, args] = mongoDBToKeyRangeArgs(key);
            return IDBKeyRange[type](...args);
        }
        return key;
    }

    var Server = function (db, name, noServerMethods) {
        var closed = false;

        this.getIndexedDB = () => db;
        this.isClosed = () => closed;

        this.add = function (table, ...args) {
            return new Promise(function (resolve, reject) {
                if (closed) {
                    reject('Database has been closed');
                    return;
                }

                var records = [];

                args.forEach(function (aip) {
                    if (isArray(aip)) {
                        records = records.concat(aip);
                    } else {
                        records.push(aip);
                    }
                });

                var transaction = db.transaction(table, transactionModes.readwrite);
                var store = transaction.objectStore(table);

                records.forEach(function (record) {
                    var req;
                    if (record.item && record.key) {
                        var key = record.key;
                        record = record.item;
                        req = store.add(record, key);
                    } else {
                        req = store.add(record);
                    }

                    req.onsuccess = function (e) {
                        var target = e.target;
                        var keyPath = target.source.keyPath;
                        if (keyPath === null) {
                            keyPath = '__id__';
                        }
                        Object.defineProperty(record, keyPath, {
                            value: target.result,
                            enumerable: true
                        });
                    };
                });

                transaction.oncomplete = () => resolve(records, this);

                transaction.onerror = e => {
                    // prevent Firefox from throwing a ConstraintError and aborting (hard)
                    // https://bugzilla.mozilla.org/show_bug.cgi?id=872873
                    e.preventDefault();
                    reject(e);
                };
                transaction.onabort = e => reject(e);
            });
        };

        this.update = function (table, ...args) {
            return new Promise(function (resolve, reject) {
                if (closed) {
                    reject('Database has been closed');
                    return;
                }

                var transaction = db.transaction(table, transactionModes.readwrite);
                var store = transaction.objectStore(table);
                var records = [];

                args.forEach(function (aip) {
                    if (isArray(aip)) {
                        records = records.concat(aip);
                    } else {
                        records.push(aip);
                    }
                });
                records.forEach(function (record) {
                    if (record.item && record.key) {
                        var key = record.key;
                        record = record.item;
                        store.put(record, key);
                    } else {
                        store.put(record);
                    }
                });

                transaction.oncomplete = () => resolve(records, this);
                transaction.onerror = e => reject(e);
                transaction.onabort = e => reject(e);
            });
        };

        this.remove = function (table, key) {
            return new Promise(function (resolve, reject) {
                if (closed) {
                    reject('Database has been closed');
                    return;
                }
                var transaction = db.transaction(table, transactionModes.readwrite);
                var store = transaction.objectStore(table);

                store.delete(key);
                transaction.oncomplete = () => resolve(key);
                transaction.onerror = e => reject(e);
                transaction.onabort = e => reject(e);
            });
        };

        this.clear = function (table) {
            return new Promise(function (resolve, reject) {
                if (closed) {
                    reject('Database has been closed');
                    return;
                }
                var transaction = db.transaction(table, transactionModes.readwrite);
                var store = transaction.objectStore(table);

                store.clear();
                transaction.oncomplete = () => resolve();
                transaction.onerror = e => reject(e);
                transaction.onabort = e => reject(e);
            });
        };

        this.close = function () {
            if (closed) {
                throw new Error('Database has been closed');
            }
            db.close();
            closed = true;
            delete dbCache[name];
        };

        this.get = function (table, key) {
            return new Promise(function (resolve, reject) {
                if (closed) {
                    reject('Database has been closed');
                    return;
                }
                var transaction = db.transaction(table);
                var store = transaction.objectStore(table);

                key = mongoifyKey(key);
                var req = store.get(key);
                req.onsuccess = e => resolve(e.target.result);
                transaction.onerror = e => reject(e);
                transaction.onabort = e => reject(e);
            });
        };

        this.query = function (table, index) {
            if (closed) {
                throw new Error('Database has been closed');
            }
            return new IndexQuery(table, db, index);
        };

        this.count = function (table, key) {
            return new Promise((resolve, reject) => {
                if (closed) {
                    reject('Database has been closed');
                    return;
                }
                var transaction = db.transaction(table);
                var store = transaction.objectStore(table);
                key = mongoifyKey(key);
                var req = key === undefined ? store.count() : store.count(key);
                req.onsuccess = e => resolve(e.target.result);
                transaction.onerror = e => reject(e);
                transaction.onabort = e => reject(e);
            });
        };

        if (noServerMethods) {
            return;
        }

        var err;
        [].some.call(db.objectStoreNames, storeName => {
            if (this[storeName]) {
                err = new Error('The store name, "' + storeName + '", which you have attempted to load, conflicts with db.js method names."');
                this.close();
                return true;
            }
            this[storeName] = {};
            var keys = Object.keys(this);
            keys.filter(key => key !== 'close')
                .map(key =>
                    this[storeName][key] = (...args) => this[key](storeName, ...args)
                );
        });
        return err;
    };

    var IndexQuery = function (table, db, indexName) {
        var modifyObj = false;

        var runQuery = function (type, args, cursorType, direction, limitRange, filters, mapper) {
            var transaction = db.transaction(table, modifyObj ? transactionModes.readwrite : transactionModes.readonly);
            var store = transaction.objectStore(table);
            var index = indexName ? store.index(indexName) : store;
            var keyRange = type ? IDBKeyRange[type](...args) : null;
            var results = [];
            var indexArgs = [keyRange];
            var counter = 0;

            limitRange = limitRange || null;
            filters = filters || [];
            if (cursorType !== 'count') {
                indexArgs.push(direction || 'next');
            }

            // create a function that will set in the modifyObj properties into
            // the passed record.
            var modifyKeys = modifyObj ? Object.keys(modifyObj) : false;

            var modifyRecord = function (record) {
                for (var i = 0; i < modifyKeys.length; i++) {
                    var key = modifyKeys[i];
                    var val = modifyObj[key];
                    if (val instanceof Function) { val = val(record); }
                    record[key] = val;
                }
                return record;
            };

            index[cursorType](...indexArgs).onsuccess = function (e) {
                var cursor = e.target.result;
                if (typeof cursor === 'number') {
                    results = cursor;
                } else if (cursor) {
                    if (limitRange !== null && limitRange[0] > counter) {
                        counter = limitRange[0];
                        cursor.advance(limitRange[0]);
                    } else if (limitRange !== null && counter >= (limitRange[0] + limitRange[1])) {
                        // out of limit range... skip
                    } else {
                        var matchFilter = true;
                        var result = 'value' in cursor ? cursor.value : cursor.key;

                        filters.forEach(function (filter) {
                            if (!filter || !filter.length) {
                                // Invalid filter do nothing
                            } else if (filter.length === 2) {
                                matchFilter = matchFilter && (result[filter[0]] === filter[1]);
                            } else {
                                matchFilter = matchFilter && filter[0](result);
                            }
                        });

                        if (matchFilter) {
                            counter++;
                            // if we're doing a modify, run it now
                            if (modifyObj) {
                                result = modifyRecord(result);
                                cursor.update(result);
                            }
                            results.push(mapper(result));
                        }
                        cursor.continue();
                    }
                }
            };

            return new Promise(function (resolve, reject) {
                transaction.oncomplete = () => resolve(results);
                transaction.onerror = e => reject(e);
                transaction.onabort = e => reject(e);
            });
        };

        var Query = function (type, args) {
            var direction = 'next';
            var cursorType = 'openCursor';
            var filters = [];
            var limitRange = null;
            var mapper = defaultMapper;
            var unique = false;

            var execute = function () {
                return runQuery(type, args, cursorType, unique ? direction + 'unique' : direction, limitRange, filters, mapper);
            };

            var limit = function (...args) {
                limitRange = args.slice(0, 2);
                if (limitRange.length === 1) {
                    limitRange.unshift(0);
                }

                return {
                    execute: execute
                };
            };
            var count = function () {
                direction = null;
                cursorType = 'count';

                return {
                    execute: execute
                };
            };

            var filter, desc, distinct, modify, map;
            var keys = function () {
                cursorType = 'openKeyCursor';

                return {
                    desc,
                    execute,
                    filter,
                    distinct,
                    map
                };
            };
            filter = function (...args) {
                filters.push(args.slice(0, 2));

                return {
                    keys,
                    execute,
                    filter,
                    desc,
                    distinct,
                    modify,
                    limit,
                    map
                };
            };
            desc = function () {
                direction = 'prev';

                return {
                    keys,
                    execute,
                    filter,
                    distinct,
                    modify,
                    map
                };
            };
            distinct = function () {
                unique = true;
                return {
                    keys,
                    count,
                    execute,
                    filter,
                    desc,
                    modify,
                    map
                };
            };
            modify = function (update) {
                modifyObj = update;
                return {
                    execute
                };
            };
            map = function (fn) {
                mapper = fn;

                return {
                    execute,
                    count,
                    keys,
                    filter,
                    desc,
                    distinct,
                    modify,
                    limit,
                    map
                };
            };

            return {
                execute,
                count,
                keys,
                filter,
                desc,
                distinct,
                modify,
                limit,
                map
            };
        };

        ['only', 'bound', 'upperBound', 'lowerBound'].forEach((name) => {
            this[name] = function () {
                return Query(name, arguments);
            };
        });

        this.range = function (opts) {
            return Query.apply(null, mongoDBToKeyRangeArgs(opts));
        };

        this.filter = function (...args) {
            var query = Query(null, null);
            return query.filter(...args);
        };

        this.all = function () {
            return this.filter();
        };
    };

    var createSchema = function (e, schema, db) {
        if (typeof schema === 'function') {
            schema = schema();
        }

        if (!schema || schema.length === 0) {
            return;
        }

        for (var objectStoreKey in db.objectStoreNames) {
            if (db.objectStoreNames.hasOwnProperty(objectStoreKey)) {
                var name = db.objectStoreNames[objectStoreKey];
                if (schema.hasOwnProperty(name) === false) {
                    e.currentTarget.transaction.db.deleteObjectStore(name);
                }
            }
        }

        var tableName;
        for (tableName in schema) {
            var table = schema[tableName];
            var store;
            if (!hasOwn.call(schema, tableName) || db.objectStoreNames.contains(tableName)) {
                store = e.currentTarget.transaction.objectStore(tableName);
            } else {
                store = db.createObjectStore(tableName, table.key);
            }

            var indexKey;
            for (indexKey in table.indexes) {
                var index = table.indexes[indexKey];
                try {
                    store.index(indexKey);
                } catch (err) {
                    store.createIndex(indexKey, index.key || indexKey, Object.keys(index).length ? index : { unique: false });
                }
            }
        }
    };

    var open = function (e, server, noServerMethods, version, schema) {
        var db = e.target.result;
        dbCache[server] = db;

        var s = new Server(db, server, noServerMethods);
        return s instanceof Error ? Promise.reject(s) : Promise.resolve(s);
    };

    var db = {
        version: '0.12.0',
        open: function (options) {
            return new Promise(function (resolve, reject) {
                if (dbCache[options.server]) {
                    open({
                        target: {
                            result: dbCache[options.server]
                        }
                    }, options.server, options.noServerMethods, options.version, options.schema)
                    .then(resolve, reject);
                } else {
                    let request = indexedDB.open(options.server, options.version);

                    request.onsuccess = e => open(e, options.server, options.noServerMethods, options.version, options.schema).then(resolve, reject);
                    request.onupgradeneeded = e => createSchema(e, options.schema, e.target.result);
                    request.onerror = e => reject(e);
                }
            });
        },

        delete: function (dbName) {
            return new Promise(function (resolve, reject) {
                var request = indexedDB.deleteDatabase(dbName);

                request.onsuccess = e => resolve(e);
                request.onerror = e => reject(e);
                request.onblocked = e => reject(e);
            });
        },

        cmp: function (param1, param2) {
            return indexedDB.cmp(param1, param2);
        }
    };

    if (typeof module !== 'undefined' && typeof module.exports !== 'undefined') {
        module.exports = db;
    } else if (typeof define === 'function' && define.amd) {
        define(function () { return db; });
    } else {
        window.db = db;
    }
})(window);
